(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Omega.
Require Import Fomega_syntax.
Require Import Fomega_vars.


(*************************************************************************)
(* Shifting *)
(*************************************************************************)

Inductive sign : Set := Up | Dn.

Definition shiftBnd i s j n :=
  if proj1_sig (Bool_nat.nat_lt_ge_bool n i) then n else
  match s with
  | Up => n + j
  | Dn => n - j
  end.

Fixpoint shiftAT i s j (t : typ) {struct t} : typ :=
  match t with
  | varT (free a) => varT (free a)
  | varT (bnd n) => varT (bnd (shiftBnd i s j n))
  | unitT => unitT
  | arrT t1 t2 => arrT (shiftAT i s j t1) (shiftAT i s j t2)
  | timT t1 t2 => timT (shiftAT i s j t1) (shiftAT i s j t2)
  | plusT t1 t2 => plusT (shiftAT i s j t1) (shiftAT i s j t2)
  | recT k1 t2 => recT k1 (shiftAT (S i) s j t2)
  | allT k1 t2 => allT k1 (shiftAT (S i) s j t2)
  | exT  k1 t2 => exT k1  (shiftAT (S i) s j t2)
  | trivT => trivT
  | lamT k1 t2 => lamT k1  (shiftAT (S i) s j t2)
  | appT t1 t2 => appT (shiftAT i s j t1) (shiftAT i s j t2)
  | pairT t1 t2 => pairT (shiftAT i s j t1) (shiftAT i s j t2)
  | fstT t1 => fstT (shiftAT i s j t1)
  | sndT t1 => sndT (shiftAT i s j t1)
  end.

Fixpoint shiftAE i s j (e : exp) {struct e} : exp :=
  match e with
  | varE x => varE x
  | trivE => trivE
  | lamE t1 e2 => lamE (shiftAT i s j t1) (shiftAE i s j e2)
  | appE e1 e2 => appE (shiftAE i s j e1) (shiftAE i s j e2)
  | pairE e1 e2 => pairE (shiftAE i s j e1) (shiftAE i s j e2)
  | fstE e1 => fstE (shiftAE i s j e1)
  | sndE e1 => sndE (shiftAE i s j e1)
  | lftE e1 t2 => lftE (shiftAE i s j e1) (shiftAT i s j t2)
  | rgtE e1 t2 => rgtE (shiftAE i s j e1) (shiftAT i s j t2)
  | caseE e1 e2 e3 => caseE (shiftAE i s j e1) (shiftAE i s j e2) (shiftAE i s j e3)
  | rollE e1 t2 => rollE (shiftAE i s j e1) (shiftAT i s j t2)
  | unrollE e1 => unrollE (shiftAE i s j e1)
  | genE k1 e2 => genE k1 (shiftAE (S i) s j e2)
  | instE e1 t2 => instE (shiftAE i s j e1) (shiftAT i s j t2)
  | packE t1 e2 t3 => packE (shiftAT i s j t1) (shiftAE i s j e2) (shiftAT i s j t3)
  | unpackE e1 e2 => unpackE (shiftAE i s j e1) (shiftAE (S i) s j e2)
  end.

Fixpoint shiftXE i s j (e : exp) {struct e} : exp :=
  match e with
  | varE (free x) => varE (free x)
  | varE (bnd n) => varE (bnd (shiftBnd i s j n))
  | trivE => trivE
  | lamE t1 e2 => lamE t1 (shiftXE (S i) s j e2)
  | appE e1 e2 => appE (shiftXE i s j e1) (shiftXE i s j e2)
  | pairE e1 e2 => pairE (shiftXE i s j e1) (shiftXE i s j e2)
  | fstE e1 => fstE (shiftXE i s j e1)
  | sndE e1 => sndE (shiftXE i s j e1)
  | lftE e1 t2 => lftE (shiftXE i s j e1) t2
  | rgtE e1 t2 => rgtE (shiftXE i s j e1) t2
  | caseE e1 e2 e3 => caseE (shiftXE i s j e1) (shiftXE (S i) s j e2) (shiftXE (S i) s j e3)
  | rollE e1 t2 => rollE (shiftXE i s j e1) t2
  | unrollE e1 => unrollE (shiftXE i s j e1)
  | genE k1 e2 => genE k1 (shiftXE i s j e2)
  | instE e1 t2 => instE (shiftXE i s j e1) t2
  | packE t1 e2 t3 => packE t1 (shiftXE i s j e2) t3
  | unpackE e1 e2 => unpackE (shiftXE i s j e1) (shiftXE (S i) s j e2)
  end.


(*************************************************************************)
(* Shifting by 0 *)
(*************************************************************************)

Lemma shiftBnd_zero : forall n s i,
  shiftBnd i s 0 n = n.
Proof.
  intros n s i.
  unfold shiftBnd.
  destruct (proj1_sig (Bool_nat.nat_lt_ge_bool n i)).
    reflexivity.
    destruct s. omega. omega.
Qed.

Lemma shiftAT_zero : forall t s i,
  shiftAT i s 0 t = t.
Proof.
  intros t s.
  induction t; intros i; simpl shiftAT; f_equal;
    try solve
    [ reflexivity
    | apply IHt
    | apply IHt1
    | apply IHt2
    ].
    Case "var".
      destruct o.
        reflexivity.
        f_equal. f_equal. apply shiftBnd_zero.
Qed.

Lemma shiftAE_zero : forall e s i,
  shiftAE i s 0 e = e.
Proof.
  intros e s.
  induction e; intros i; simpl shiftAE; f_equal;
    try solve
    [ reflexivity
    | apply IHe
    | apply IHe1
    | apply IHe2
    | apply IHe3
    | apply shiftAT_zero
    ].
Qed.

Lemma shiftXE_zero : forall e s i,
  shiftXE i s 0 e = e.
Proof.
  intros e s.
  induction e; intros i; simpl shiftXE; f_equal;
    try solve
    [ reflexivity
    | apply IHe
    | apply IHe1
    | apply IHe2
    | apply IHe3
    ].
    Case "var".
      destruct o.
        reflexivity.
        f_equal. f_equal. apply shiftBnd_zero.
Qed.


(*************************************************************************)
(* Facts about comparison *)
(*************************************************************************)

Lemma lt_less : forall i j,
  i < j ->
  proj1_sig (Bool_nat.nat_lt_ge_bool i j) = true.
Proof.
  intros i j Lt.
  destruct (Bool_nat.nat_lt_ge_bool i j) as [b Ge].
  destruct b.
    Case "true".
      reflexivity.
    Case "false".
      absurd (j < i /\ j >= i). omega. omega.
Qed.

Lemma lt_greatereq : forall i j,
  i >= j ->
  proj1_sig (Bool_nat.nat_lt_ge_bool i j) = false.
Proof.
  intros i j Lt.
  destruct (Bool_nat.nat_lt_ge_bool i j) as [b Ge].
  destruct b.
    Case "true".
      absurd (j < i /\ j >= i). omega. omega.
    Case "false".
      reflexivity.
Qed.


Lemma ge_plus : forall n m i,
  proj1_sig (Bool_nat.nat_lt_ge_bool n m) = false ->
  proj1_sig (Bool_nat.nat_lt_ge_bool (n + i) m) = false.
Proof.
  intros n m i H.
  destruct (Bool_nat.nat_lt_ge_bool n m).
  unfold proj1_sig in H.
  rewrite H in y.
  assert (H' : n + i >= m). omega.
  destruct (Bool_nat.nat_lt_ge_bool (n + i) m).
  unfold proj1_sig.
  destruct x0.
    unfold ge, lt in *.
      (* there must be an easier way... *)
      assert (A : S (n + i) <= n + i). omega. 
      assert (T : n + i < S (n + i)). omega.
      assert (AA : False). apply (lt_not_le (n + i) (S (n + i)) T A).
      inversion AA.
    reflexivity.
Qed.


(*************************************************************************)
(* Shifting closed terms *)
(*************************************************************************)

Lemma shiftBnd_irrel : forall i s j n,
  n < i ->
  shiftBnd i s j n = n.
Proof.
  intros i s j n Less.
  unfold shiftBnd.
  rewrite lt_less; simpl.
    reflexivity.
    inversion Less. omega. omega. 
Qed.


Lemma shiftAT_substAT_irrel : forall i s j t (n : nat) t',
  n < i ->
  substAT n t' t = shiftAT i s j (substAT n t' t) ->
  shiftAT i s j t = t.
Proof.
  intros i s j t n t'.
  unfold substAT, shiftAT.
  generalize dependent n. generalize dependent i.
  induction t; intros i n Hn H; inversion H; fold substAT in *; fold shiftAT in *; simpl; f_equal;
    try solve
    [ reflexivity
    | first [eapply IHt with (n := n) | eapply IHt1 with (n := n)]; [omega | apply H1]
    | eapply IHt2 with (n := n); [omega | apply H2]
    | eapply IHt with (n := S n); [omega | apply H1]
    ].
    Case "var".
      destruct o.
        reflexivity.
        destruct (eqOcc n0 n).
          f_equal. f_equal. apply shiftBnd_irrel. inversion e. omega.
          symmetry. apply H.
Qed.

Lemma shiftAT_lclosed : forall i s j t,
  lclosedT t ->
  shiftAT i s j t = t.
Proof.
  unfold shiftAT.
  intros i s j t Ht. generalize dependent i.
  induction Ht; intros i; fold shiftAT in *; f_equal;
    try solve
    [ reflexivity
    | apply IHHt
    | apply IHHt1
    | apply IHHt2
    | unfold openAT in *; pick_fresh_gen L a; f_equal;
        eapply (shiftAT_substAT_irrel (S i) s j t2 0);
          [omega | symmetry; apply H0; apply Fr]
    ].
Qed.


Lemma shiftAE_substAE_irrel : forall i s j e (n : nat) t,
  n < i ->
  substAE n t e = shiftAE i s j (substAE n t e) ->
  shiftAE i s j e = e.
Proof.
  intros i s j e n t.
  unfold substAE, shiftAE.
  generalize dependent n. generalize dependent i.
  induction e; intros i n Hn H; inversion H; fold substAE in *; fold shiftAE in *; simpl; f_equal;
    try solve
    [ reflexivity
    | first [eapply IHe with (n := n) | eapply IHe1 with (n := n)]; [omega | apply H1]
    | eapply IHe2 with (n := n); [omega | apply H2]
    | eapply IHe3 with (n := n); [omega | apply H3]
    | eapply IHe with (n := S n); [omega | apply H1]
    | eapply IHe2 with (n := S n); [omega | apply H2]
    | eapply IHe; [apply Hn | first [apply H1 | apply H2]]
    | eapply shiftAT_substAT_irrel; [apply Hn | first [apply H1 | apply H2 | apply H3]]
    ].
Qed.

Lemma shiftXE_substAE_irrel : forall i s j e a t,
  substAE a t e = shiftXE i s j (substAE a t e) ->
  shiftXE i s j e = e.
Proof.
  intros i s j e a t.
  unfold shiftXE, substAE.
  generalize dependent a. generalize dependent i.
  induction e; intros i a H; inversion H; fold shiftXE in *; fold substAE in *; f_equal;
    try solve
    [ reflexivity
    | eapply IHe; apply H1
    | eapply IHe1; apply H1
    | eapply IHe2; apply H2
    | eapply IHe3; apply H3
    ].
Qed.

Lemma shiftAE_substXE_irrel : forall i s j e x e',
  lclosedE e' ->
  substXE x e' e = shiftAE i s j (substXE x e' e) ->
  shiftAE i s j e = e.
Proof.
  intros i s j e x e' He'.
  unfold shiftAE, substXE.
  generalize dependent x. generalize dependent i.
  induction e; intros i x H; inversion H; fold shiftAE in *; fold substXE in *; f_equal;
    try solve
    [ reflexivity
    | eapply IHe; first [apply H1 | apply H2 | apply H3]
    | eapply IHe1; apply H1
    | eapply IHe2; apply H2
    | eapply IHe3; apply H3
    | rewrite <- H1; symmetry; apply H1
    | rewrite <- H2; symmetry; apply H2
    | rewrite <- H3; symmetry; apply H3
    ].
Qed.

Lemma shiftAE_lclosed : forall i s j e,
  lclosedE e ->
  shiftAE i s j e = e.
Proof.
  unfold shiftAE.
  intros i s j e He. generalize dependent i.
  induction He; intros i; fold shiftAE in *; f_equal;
    try solve
    [ reflexivity
    | apply IHHe
    | apply IHHe1
    | apply IHHe2
    | eapply shiftAT_lclosed; first [apply H | apply H0]
    | unfold openXE in *; pick_fresh_gen L x; eapply shiftAE_substXE_irrel;
        [ apply lclosedE_var
        | symmetry; first [apply H0 | apply H1 | apply H2]; apply Fr ]
    ].
    Case "gen".
        unfold openAE in *.
        pick_fresh_gen L a.
        eapply shiftAE_substAE_irrel with (n := 0).
          omega.
          symmetry. apply H0. apply Fr .
    Case "open 2".
      unfold openXE, openAE in *.
      pick_fresh_gen L a.
      pick_fresh_gen L x.
      eapply shiftAE_substAE_irrel with (n := 0). omega.
      symmetry.
      eapply shiftAE_substXE_irrel. apply lclosedE_var.
      symmetry. apply H0. apply Fr. apply Fr0.
Qed.


Lemma shiftXE_substXE_irrel : forall i s j e (n : nat) e',
  n < i ->
  substXE n e' e = shiftXE i s j (substXE n e' e) ->
  shiftXE i s j e = e.
Proof.
  intros i s j e n e'.
  unfold substXE, shiftXE.
  generalize dependent n. generalize dependent i.
  induction e; intros i n Hn H; inversion H; fold substXE in *; fold shiftXE in *; simpl; f_equal;
    try solve
    [ reflexivity
    | eapply IHe with (n := n); [omega | apply H1]
    | eapply IHe1 with (n := n); [omega | apply H1]
    | eapply IHe2 with (n := n); [omega | apply H2]
    | eapply IHe with (n := S n); [omega | apply H1]
    | eapply IHe2 with (n := S n); [omega | apply H2]
    | eapply IHe3 with (n := S n); [omega | apply H3]
    ].
    Case "var".
      destruct o.
        reflexivity.
        destruct (eqOcc n0 n).
          f_equal. f_equal. apply shiftBnd_irrel. inversion e. omega.
          symmetry. apply H.
Qed.

Lemma shiftXE_lclosed : forall i s j e,
  lclosedE e ->
  shiftXE i s j e = e.
Proof.
  unfold shiftXE.
  intros i s j e He. generalize dependent i.
  induction He; intros i; fold shiftXE in *; f_equal;
    try solve
    [ reflexivity
    | apply IHHe
    | apply IHHe1
    | apply IHHe2
    | unfold openXE in *; pick_fresh_gen L x; eapply shiftXE_substXE_irrel with (n := 0);
        [ omega
        | symmetry; first [apply H0 | apply H1 | apply H2]; apply Fr ]
    ].
    Case "gen".
      unfold openAE in *.
      pick_fresh_gen L a.
      eapply shiftXE_substAE_irrel. symmetry. apply H0. apply Fr.
    Case "open 2".
      unfold openXE in *.
      pick_fresh_gen L a.
      pick_fresh_gen L x.
      eapply shiftXE_substAE_irrel.
      symmetry.
      eapply shiftXE_substXE_irrel with (n := 0). omega.
      symmetry. unfold openAE in H0. apply H0. apply Fr. apply Fr0.
Qed.


(*************************************************************************)
(* Nested shifting *)
(*************************************************************************)

Lemma shiftAT_shiftAT : forall t i j1 j2,
  shiftAT i Up j1 (shiftAT i Up j2 t) = shiftAT i Up (j1 + j2) t.
Proof.
  intros t i j1 j2.
  unfold shiftAT.
  generalize dependent i.
  induction t; intros i; f_equal;
    try solve
    [ reflexivity
    | apply IHt
    | apply IHt1
    | apply IHt2
    ].
    Case "var".
      destruct o.
        SCase "free".
          reflexivity.
        SCase "bound".
          unfold shiftBnd.
          case_eq (proj1_sig (Bool_nat.nat_lt_ge_bool n i)); intros B.
            rewrite B. reflexivity.
          case_eq (proj1_sig (Bool_nat.nat_lt_ge_bool (n + j2) i)); intros B'.
            rewrite ge_plus in B'. inversion B'. apply B.
            f_equal. f_equal. omega.
Qed.

Lemma shiftAE_shiftAE : forall e i j1 j2,
  shiftAE i Up j1 (shiftAE i Up j2 e) = shiftAE i Up (j1 + j2) e.
Proof.
  intros e i j1 j2.
  generalize dependent i.
  induction e; intros i; simpl shiftAE; f_equal;
    try solve
    [ reflexivity
    | apply IHe
    | apply IHe1
    | apply IHe2
    | apply IHe3
    | apply shiftAT_shiftAT
    ].
Qed.

Lemma shiftXE_shiftXE : forall e i j1 j2,
  shiftXE i Up j1 (shiftXE i Up j2 e) = shiftXE i Up (j1 + j2) e.
Proof.
  intros e i j1 j2.
  unfold shiftXE.
  generalize dependent i.
  induction e; intros i; f_equal;
    try solve
    [ reflexivity
    | apply IHe
    | apply IHe1
    | apply IHe2
    | apply IHe3
    ].
    Case "var".
      destruct o.
        SCase "free".
          reflexivity.
        SCase "bound".
          unfold shiftBnd.
          case_eq (proj1_sig (Bool_nat.nat_lt_ge_bool n i)); intros B.
            rewrite B. reflexivity.
          case_eq (proj1_sig (Bool_nat.nat_lt_ge_bool (n + j2) i)); intros B'.
            rewrite ge_plus in B'. inversion B'. apply B.
            f_equal. f_equal. omega.
Qed.


(*************************************************************************)
(* Commuting shifting *)
(*************************************************************************)

Lemma shiftAT_substAT_free : forall t i s j (a : var) t',
  lclosedT t' ->
  shiftAT i s j (substAT a t' t) = substAT a t' (shiftAT i s j t).
Proof.
  intros t i s j a t' Ht'. generalize dependent i.
  induction t; intros i; simpl shiftAT in *; simpl substAT in *; simpl shift; f_equal;
    try solve
    [ reflexivity
    | apply IHt; omega
    | apply IHt1; omega
    | apply IHt2; omega
    ].
    Case "var".
      destruct o; subst; unfold eqOcc.
        SCase "o = x".
          destruct (v == a); subst.
            rewrite shiftAT_lclosed. 2: apply Ht'.
              simpl substAT. destruct (a == a).
                reflexivity.
                destruct n. reflexivity.
            rewrite shiftAT_lclosed. 2: apply lclosedT_var.
              simpl substAT. destruct (v == a).
                destruct (n e).
                reflexivity.
        SCase "o = n".
          simpl shiftAT. simpl substAT. reflexivity.
Qed.

Lemma substAT_shiftAT_irrel : forall t (i : nat) j t',
  j <> 0 ->
  substAT i t' (shiftAT i Up j t) = shiftAT i Up j t.
Proof.
  intros t i j t' Hj.
  unfold shiftAT, substAT.
  generalize dependent i.
  induction t; intro i; fold shiftAT in *; fold substAT in *; unfold shift; f_equal;
    try solve
    [ reflexivity
    | try rewrite <- plus_Sn_m; first [apply IHt | apply IHt1 | apply IHt2]; omega
    ].
    Case "var".
      destruct o; subst; unfold eqOcc.
        SCase "o = a".
          unfold substAT, eqOcc. reflexivity.
        SCase "o = n".
          unfold shiftBnd.
          destruct (Bool_nat.nat_lt_ge_bool n i) as [b LGE]. destruct b; simpl.
            destruct (n === i).
              absurd (n = i /\ n < i). omega. omega.
              reflexivity.
            destruct (n + j === i).
              absurd (n + j = i /\ n > i). omega. omega.
              reflexivity.
Qed.

Lemma shiftAT_substAT_lt : forall t i j (n : nat) t',
  n < i ->
  lclosedT t' ->
  shiftAT i Up j (substAT n t' t) = substAT n t' (shiftAT i Up j t).
Proof.
  intros t i j n t' Hn Ht'.
  unfold shiftAT, substAT.
  generalize dependent i. generalize dependent n.
  induction t; intros n i Hn; fold shiftAT in *; fold substAT in *; unfold shift; f_equal;
    try solve
    [ reflexivity
    | apply IHt; omega
    | apply IHt1; omega
    | apply IHt2; omega
    ].
    Case "var".
      destruct o; subst; unfold eqOcc.
        SCase "o = x".
          rewrite shiftAT_lclosed.
            unfold substXE, eqOcc. destruct (n === n).
              reflexivity.
              absurd (n = n /\ n <> n). omega. omega.
            apply lclosedT_var.
        SCase "o = n".
          destruct (n0 === n); subst.
            rewrite shiftAT_lclosed.
              unfold shiftBnd.
              rewrite lt_less.
                unfold substAT, eqOcc. destruct (n === n).
                  reflexivity.
                  destruct n0. reflexivity.
                apply Hn.
              apply Ht'.
            unfold shiftAT. unfold shiftBnd.
              destruct (Bool_nat.nat_lt_ge_bool n0 i) as [b LGE]. destruct b; simpl.
                destruct (n0 === n).
                  absurd (n0 = n /\ n0 <> n). omega. omega.
                  reflexivity.
                destruct (n0 + j === n).
                  absurd (n0 + j = n /\ n0 + j > n). omega. omega.
                  reflexivity.
Qed.

Lemma shiftAT_substAT_ge : forall t i j (n : nat) t',
  n >= i ->
  lclosedT t' ->
  shiftAT i Up j (substAT n t' t) = substAT (n + j) t' (shiftAT i Up j t).
Proof.
  intros t i j n t' Hn Ht'.
  unfold shiftAT, substAT.
  generalize dependent i. generalize dependent n.
  induction t; intros n i Hn; fold shiftAT in *; fold substAT in *; unfold shift; f_equal;
    try solve
    [ reflexivity
    | try rewrite <- plus_Sn_m; first [apply IHt | apply IHt1 | apply IHt2]; omega
    ].
    Case "var".
      destruct o; subst; unfold eqOcc.
        SCase "o = a".
          rewrite shiftAT_lclosed.
            unfold substAT, eqOcc. reflexivity.
            apply lclosedT_var.
        SCase "o = n".
          destruct (n0 === n); subst.
            rewrite shiftAT_lclosed.
            unfold shiftBnd.
              rewrite lt_greatereq.
                unfold substAT, eqOcc. destruct (n + j === n + j).
                  reflexivity.
                  destruct n0. reflexivity.
                apply Hn.
              apply Ht'.
            unfold shiftAT.
            unfold shiftBnd.
              destruct (Bool_nat.nat_lt_ge_bool n0 i) as [b LGE]. destruct b; simpl.
                destruct (n0 === n + j).
                  absurd (n0 = n + j /\ n0 <> n). omega. omega.
                  reflexivity.
                destruct (n0 + j === n + j).
                  absurd (n0 + j = n /\ n0 + j > n). omega. omega.
                  reflexivity.
Qed.


Lemma shiftAE_substXE : forall e i j a t,
  shiftXE i Up j (substAE a t e) = substAE a t (shiftXE i Up j e).
Proof.
  intros e i j n t.
  unfold shiftXE, substAE.
  generalize dependent i. generalize dependent n.
  induction e; intros n i; fold shiftXE in *; fold substAE in *; f_equal;
    try solve
    [ reflexivity
    | apply IHe
    | apply IHe1
    | apply IHe2
    | apply IHe3
    ].
    Case "var".
      destruct o; reflexivity.
Qed.


Lemma substXE_shiftXE_irrel : forall e (i : nat) j e',
  j <> 0 ->
  substXE i e' (shiftXE i Up j e) = shiftXE i Up j e.
Proof.
  intros e i j e' Hj.
  unfold shiftXE, substXE.
  generalize dependent i.
  induction e; intro i; fold shiftXE in *; fold substXE in *; unfold shift; f_equal;
    try solve
    [ reflexivity
    | try rewrite <- plus_Sn_m; first [apply IHe | apply IHe1 | apply IHe2 | apply IHe3]; omega
    ].
    Case "var".
      destruct o; subst; unfold eqOcc.
        SCase "o = x".
          unfold substXE, eqOcc. reflexivity.
        SCase "o = n".
          unfold shiftBnd.
          destruct (Bool_nat.nat_lt_ge_bool n i) as [b LGE]. destruct b; simpl.
            destruct (n === i).
              absurd (n = i /\ n < i). omega. omega.
              reflexivity.
            destruct (n + j === i).
              absurd (n + j = i /\ n > i). omega. omega.
              reflexivity.
Qed.

Lemma shiftXE_substXE_lt : forall e i j (n : nat) e',
  n < i ->
  lclosedE e' ->
  shiftXE i Up j (substXE n e' e) = substXE n e' (shiftXE i Up j e).
Proof.
  intros e i j n e' Hn He'.
  unfold shiftXE, substXE.
  generalize dependent i. generalize dependent n.
  induction e; intros n i Hn; fold shiftXE in *; fold substXE in *; unfold shift; f_equal;
    try solve
    [ reflexivity
    | apply IHe; omega
    | apply IHe1; omega
    | apply IHe2; omega
    | apply IHe3; omega
    ].
    Case "var".
      destruct o; subst; unfold eqOcc.
        SCase "o = x".
          rewrite shiftXE_lclosed.
            unfold substXE, eqOcc. destruct (n === n).
              reflexivity.
              absurd (n = n /\ n <> n). omega. omega.
            apply lclosedE_var.
        SCase "o = n".
          destruct (n0 === n); subst.
            rewrite shiftXE_lclosed.
              unfold shiftBnd.
              rewrite lt_less.
                unfold substXE, eqOcc. destruct (n === n).
                  reflexivity.
                  destruct n0. reflexivity.
                apply Hn.
              apply He'.
            unfold shiftXE. unfold shiftBnd.
              destruct (Bool_nat.nat_lt_ge_bool n0 i) as [b LGE]. destruct b; simpl.
                destruct (n0 === n).
                  absurd (n0 = n /\ n0 <> n). omega. omega.
                  reflexivity.
                destruct (n0 + j === n).
                  absurd (n0 + j = n /\ n0 + j > n). omega. omega.
                  reflexivity.
Qed.

Lemma shiftXE_substXE_ge : forall e i j (n : nat) e',
  n >= i ->
  lclosedE e' ->
  shiftXE i Up j (substXE n e' e) = substXE (n + j) e' (shiftXE i Up j e).
Proof.
  intros e i j n e' Hn He'.
  unfold shiftXE, substXE.
  generalize dependent i. generalize dependent n.
  induction e; intros n i Hn; fold shiftXE in *; fold substXE in *; unfold shift; f_equal;
    try solve
    [ reflexivity
    | try rewrite <- plus_Sn_m; first [apply IHe | apply IHe1 | apply IHe2 | apply IHe3]; omega
    ].
    Case "var".
      destruct o; subst; unfold eqOcc.
        SCase "o = x".
          rewrite shiftXE_lclosed.
            unfold substXE, eqOcc. reflexivity.
            apply lclosedE_var.
        SCase "o = n".
          unfold shiftBnd.
          destruct (n0 === n); subst.
            rewrite shiftXE_lclosed.
              rewrite lt_greatereq.
                unfold substXE, eqOcc. destruct (n + j === n + j).
                  reflexivity.
                  destruct n0. reflexivity.
                apply Hn.
              apply He'.
            unfold shiftXE. unfold shiftBnd.
              destruct (Bool_nat.nat_lt_ge_bool n0 i) as [b LGE]. destruct b; simpl.
                destruct (n0 === n + j).
                  absurd (n0 = n + j /\ n0 <> n). omega. omega.
                  reflexivity.
                destruct (n0 + j === n + j).
                  absurd (n0 + j = n /\ n0 + j > n). omega. omega.
                  reflexivity.
Qed.
